#include "dxstdafx.h"
#include "astar.h"
#include <vector>

AStar::AStar(D3DXVECTOR3 start, D3DXVECTOR3 goal, Terrain* t, float hw)
{
	int sr, sc, gr, gc;
	t->GetRowColumn( &start, &sr, &sc );
	t->GetRowColumn( &goal, &gr, &gc );
	this->start.x = sr;
	this->start.y = sc;
	this->goal.x = gr;
	this->goal.y = gc;
	Node temp;
	temp.parent = NULL;
	temp.pos = this->start;
	temp.h = temp.f = CalculateH(sr, sc);
	temp.g = 0;
	openList.push_back(temp);
	worldWidth = TERRAIN_WIDTH - 1;
	worldHeight = TERRAIN_WIDTH - 1;
	terrain = t;
	heuristicWeight = hw;
}

AStar::~AStar()
{
}
/*
	return codes:
		-1=no path found
		0=end of path loop
		1=path found
*/
int AStar::Update()
{
	std::list<Node>::iterator iter, itemp;
	Node* curNode;

	//find lowest cost node on the open list and pop it
	curNode = FindLowestOpen();
	//no path found
	if(curNode == NULL)
		return(-1);
	//push the node from the open list onto the closed list
	closedList.push_back(*curNode);
	//erase the node from the openlist
	for(iter = openList.begin(); iter != openList.end(); ++iter)
	{
		//get the iter to the specified spot as to erase it
		if(*curNode == *iter)
		{
			openList.erase(iter);
			break;
		}
	}
	//set the index to the tiles new location on the closed list
	for(iter = closedList.begin(); iter != closedList.end(); ++iter)
	{
		itemp = iter;
	}
	curNode = &(*itemp);
	if(curNode->pos == goal)
		return(1);
	//now look at the nodes around the closed and update the open list with them
	UpdateOpenList(curNode);
	return(0);
}

//void AStar::UpdateOpenList(int curIndex)
void AStar::UpdateOpenList(Node* curNode)
{
	int x, y, prevIndex=-1;
	Node nodeTemp;
	Node *prevNode=NULL;
	for(x = -1; x <= 1; ++x)
	{
		for(y = -1; y <= 1; ++y)
		{
			if(x == 0 && y == 0)
				continue;
			nodeTemp.pos.x = curNode->pos.x + x;
			nodeTemp.pos.y = curNode->pos.y + y;
			//make sure that it is in the world
			if(nodeTemp.pos.x < 0 || nodeTemp.pos.x > worldWidth)
				continue;
			else if(nodeTemp.pos.y < 0 || nodeTemp.pos.y > worldHeight)
				continue;
			//make sure that it is not a wall
			if(terrain->IsWall(nodeTemp.pos.x, nodeTemp.pos.y) == true)
				continue;
			nodeTemp.parent = curNode;
			//update the g cost based on it being a diagonal or not
			if(x == 0 || y == 0)
				nodeTemp.g = curNode->g + 1;
			else
			{
				nodeTemp.g = curNode->g + 1.4f;
				//make sure that a corner would not be cut using this
				if((terrain->IsWall(curNode->pos.x + x, curNode->pos.y) == true) ||
					(terrain->IsWall(curNode->pos.x, curNode->pos.y + y) == true))
					continue;
			}
			nodeTemp.h = CalculateH(nodeTemp.pos.x, nodeTemp.pos.y);
			nodeTemp.f = nodeTemp.g + (nodeTemp.h * heuristicWeight);
			if(CheckIfOnClosed(nodeTemp.pos) == true)
				continue;
			prevNode=CheckIfOnOpen(nodeTemp.pos);
			if(prevNode != NULL)
			{
				if(nodeTemp.f < prevNode->f)
					*prevNode = nodeTemp;
				continue;
			}
			else//not on the open list
			{
				openList.push_back(nodeTemp);
			}
		}
	}
}
/*
Calculates the estimated cost
*/
float AStar::CalculateH(int x, int y)
{
	Coord tempspot;
	tempspot.x = x;
	tempspot.y = y;
	int xdiff = goal.x - x;
	int ydiff = goal.y - y;

	
	int diff;
	float hcost=0;
	
	if(abs(xdiff) < abs(ydiff))
	{
		tempspot.x += xdiff;
		if(ydiff > 0)
			tempspot.y += abs(xdiff);
		else
			tempspot.y -= abs(xdiff);
		hcost += abs(xdiff) * 1.4f;
		diff = goal.y - tempspot.y;
		hcost += abs(diff);
	}
	else
	{
		tempspot.y += ydiff;
		if(xdiff > 0)
			tempspot.x += abs(ydiff);
		else
			tempspot.x -= abs(ydiff);
		hcost += abs(ydiff) * 1.4f;
		diff = goal.x - tempspot.x;
		hcost += abs(diff);
	}
	return(hcost);
}

Node* AStar::FindLowestOpen()
{
	std::list<Node>::iterator iter;
	unsigned int count=0;
	Node *ret=NULL;
	float lowest=pow(2.0f, 16.0f);

	for(iter = openList.begin(); iter != openList.end(); ++iter)
	{
		if(iter->f < lowest)
		{
			ret = &(*iter);
			lowest = iter->f;
		}
	}
	//return(index);
	return(ret);
}
/*
Returns -1 if the point is not on the open list.  Otherwise it returns the index of the point
*/
Node* AStar::CheckIfOnOpen(Coord point)
{
	//int counter = 0;
	for(std::list<Node>::iterator iter=openList.begin(); iter != openList.end(); ++iter)
	{
		if(iter->pos == point)
			return(&(*iter));
	}
	return(NULL);
}
/*
Returns true if the point is on the closed list, false otherwise
*/
bool AStar::CheckIfOnClosed(Coord point)
{
	for(std::list<Node>::iterator iter=closedList.begin(); iter != closedList.end(); ++iter)
	{
		if(iter->pos == point)
			return(true);
	}
	return(false);
}

bool operator==(Node n1, Node n2)
{
	if(n1.f != n2.f || n1.g != n2.g || n1.h != n2.h)
		return(false);
	if(n1.pos != n2.pos)
		return(false);
	return(true);
}

bool operator==(Coord t, Coord t2)
{
	if(t.x == t2.x && t.y == t2.y)
		return(true);
	else
		return(false);
}

bool operator!=(Coord t, Coord t2)
{
	if(t.x != t2.x || t.y != t2.y)
		return(true);
	else
		return(false);
}