/* Copyright Steve Rabin, 2005. 
 * All rights reserved worldwide.
 *
 * This software is provided "as is" without express or implied
 * warranties. You may freely copy and compile this source into
 * applications you distribute provided that the copyright text
 * below is included in the resulting source code, for example:
 * "Portions Copyright Steve Rabin, 2005"
 */

#include "dxstdafx.h"
#include "gameobject.h"
#include "msgroute.h"
#include "statemch.h"
#include "movement.h"
#include "body.h"


GameObject::GameObject( objectID id, unsigned int type, char* name )
: m_markedForDeletion(false),
  m_stateMachineChange(NO_STATE_MACHINE_CHANGE),
  m_newStateMachine(0),
  m_body(0),
  m_movement(0),
  m_tiny(0),
  m_alive(true)
{
	m_id = id;
	m_type = type;
	
	if( strlen(name) < GAME_OBJECT_MAX_NAME_SIZE ) {
		strcpy( m_name, name );
	}
	else {
		strcpy( m_name, "invalid_name" );
		ASSERTMSG(0, "GameObject::GameObject - name is too long" );
	}
}

GameObject::~GameObject( void )
{
	DestroyStateMachineList();
}

void GameObject::CreateMovement( void )
{
	m_movement = new Movement( *this ); 
}

void GameObject::CreateBody( int health, D3DXVECTOR3& pos )
{
	m_body = new Body( health, pos, *this );
}

void GameObject::CreateTiny( CMultiAnim *pMA, std::vector< CTiny* > *pv_pChars, CSoundManager *pSM, double dTimeCurrent )
{
    m_tiny = new CTiny( *this );
    if( m_tiny == NULL )
	{
		return;
	}

    if( SUCCEEDED( m_tiny->Setup( pMA, pv_pChars, pSM, dTimeCurrent ) ) )
	{
        m_tiny->SetSounds( true );
	}
	else
	{
		delete m_tiny;
	}
}

void GameObject::Initialize( void )
{

}

void GameObject::SetName(char* name)					
{
	if( strlen(name) < GAME_OBJECT_MAX_NAME_SIZE ) 
		strcpy( m_name, name );
}

/*---------------------------------------------------------------------------*
  Name:         Update

  Description:  Calls the update function of the currect state machine.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::Update( void )
{
	if( !m_stateMachineList.empty() ) {
		ProcessStateMachineChangeRequests();
		m_stateMachineList.back()->Update();
	}
}

void GameObject::Animate( double dTimeDelta )
{
	if( m_movement )
	{
		m_movement->Animate( dTimeDelta );
	}
	if( m_tiny )
	{
		//m_tiny->Animate( dTimeDelta );
	}
}

void GameObject::AdvanceTime( double dTimeDelta, D3DXVECTOR3 *pvEye )
{
	if( m_tiny )
	{
		m_tiny->AdvanceTime( dTimeDelta, pvEye );
	}
}

void GameObject::Draw( IDirect3DDevice9* pd3dDevice, D3DXMATRIX* pViewProj )
{
	if( m_tiny )
	{
		m_tiny->Draw();
	}
	if( m_movement )
	{
		m_movement->DrawDebugVisualization( pd3dDevice, pViewProj );
	}
}

void GameObject::RestoreDeviceObjects( LPDIRECT3DDEVICE9 pd3dDevice )
{
	if( m_tiny )
	{
		m_tiny->RestoreDeviceObjects( pd3dDevice );
	}
}

void GameObject::InvalidateDeviceObjects( void )
{

}

/*---------------------------------------------------------------------------*
  Name:         SendMsg

  Description:  Sends a message to the currently active state machine.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::SendMsg( MSG_Object msg )
{
	if( !m_stateMachineList.empty() ) {
		m_stateMachineList.back()->Process( EVENT_Message, &msg );
	}
}

/*---------------------------------------------------------------------------*
  Name:         ProcessStateMachineChangeRequests

  Description:  Checks if a state machine should be changed. It will loop
                until no more state machine change requests have been made.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::ProcessStateMachineChangeRequests( void )
{
	int safetyCount = 20;
	StateMachineChange change = m_stateMachineChange;
	StateMachine * tempStateMachine = m_newStateMachine;
	
	while( change != NO_STATE_MACHINE_CHANGE && (--safetyCount >= 0) )
	{
		ASSERTMSG( safetyCount > 0, "GameObject::ProcessStateMachineChangeRequests - State machiens are flip-flopping in an infinite loop." );

		m_stateMachineChange = NO_STATE_MACHINE_CHANGE;			//Reset
		m_newStateMachine = 0;									//Reset

		switch( change )
		{
			case STATE_MACHINE_RESET:
				g_msgroute.PurgeScopedMsg( m_id ); //Remove all delayed messages addressed to me that are scoped
				ResetStateMachine();
				break;
				
			case STATE_MACHINE_REPLACE:
				g_msgroute.PurgeScopedMsg( m_id ); //Remove all delayed messages addressed to me that are scoped
				ReplaceStateMachine( *tempStateMachine );
				break;
				
			case STATE_MACHINE_QUEUE:
				QueueStateMachine( *tempStateMachine );
				break;
				
			case STATE_MACHINE_REQUEUE:
				g_msgroute.PurgeScopedMsg( m_id ); //Remove all delayed messages addressed to me that are scoped
				RequeueStateMachine();
				break;
				
			case STATE_MACHINE_PUSH:
				g_msgroute.PurgeScopedMsg( m_id ); //Remove all delayed messages addressed to me that are scoped
				PushStateMachine( *tempStateMachine, true );
				break;
				
			case STATE_MACHINE_POP:
				g_msgroute.PurgeScopedMsg( m_id ); //Remove all delayed messages addressed to me that are scoped
				PopStateMachine();
				break;
				
			default:
				ASSERTMSG( 0, "GameObject::ProcessStateMachineChangeRequests - invalid StateMachineChange request." );
		}
				
		//Check if new change
		change = m_stateMachineChange;
		tempStateMachine = m_newStateMachine;
	}
}

/*---------------------------------------------------------------------------*
  Name:         RequestStateMachineChange

  Description:  Requests that a state machine change take place.

  Arguments:    mch    : the new state machine
                change : the change to take place

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::RequestStateMachineChange( StateMachine * mch, StateMachineChange change )
{
	ASSERTMSG( m_stateMachineChange == NO_STATE_MACHINE_CHANGE, "GameObject::RequestStateMachineChange - Change already requested." );
	m_newStateMachine = mch;
	m_stateMachineChange = change;
}

/*---------------------------------------------------------------------------*
  Name:         ResetStateMachine

  Description:  Resets the current state machine.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::ResetStateMachine( void )
{
	ASSERTMSG( m_stateMachineList.size() > 0, "GameObject::ResetStateMachine - No existing state machine to reset." );
	if( m_stateMachineList.size() > 0 ) {
		StateMachine * mch = m_stateMachineList.back();
		mch->Reset();
	}
}

/*---------------------------------------------------------------------------*
  Name:         ReplaceStateMachine

  Description:  Replaces the current state machine with the provided one
                by popping off the current state machine and pushing the 
				new one.

  Arguments:    mch : the new state machine

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::ReplaceStateMachine( StateMachine & mch )
{
	ASSERTMSG( m_stateMachineList.size() > 0, "GameObject::ReplaceStateMachine - No existing state machine to replace." );
	if( m_stateMachineList.size() > 0 ) {
		StateMachine * temp = m_stateMachineList.back();
		m_stateMachineList.pop_back();
		delete( temp );
	}
	PushStateMachine( mch, true );
}

/*---------------------------------------------------------------------------*
  Name:         QueueStateMachine

  Description:  Queues a state machine behind all others, except for the
                very last one. The last state machine is the "default" and
				should always remain the last.

  Arguments:    mch : the new state machine

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::QueueStateMachine( StateMachine & mch )
{
	//Insert state machine one up from bottom
	if( m_stateMachineList.size() <= 1 ) {
		PushStateMachine( mch, false );
	}
	else {
		stateMachineListContainer::iterator i;
		i = m_stateMachineList.begin();
		i++;	//Move iterator past the first entry
		m_stateMachineList.insert(i, &mch);
		//Purposely do not "Reset" state machine until it is active
	}
}

/*---------------------------------------------------------------------------*
  Name:         RequeueStateMachine

  Description:  Requeues the current state machine behind all others, except
                for the very last one.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::RequeueStateMachine( void )
{
	ASSERTMSG( m_stateMachineList.size() > 0, "GameObject::RequeueStateMachine - No existing state machines to requeue." );
	if( m_stateMachineList.size() > 1 ) {
		StateMachine * mch = m_stateMachineList.back();
		QueueStateMachine( *mch );
		m_stateMachineList.pop_back();

		//Initialize new state machine
		mch = m_stateMachineList.back();
		mch->Reset();
	}
	else if( m_stateMachineList.size() == 1 ) {
		//Just reinitialize
		StateMachine * mch = m_stateMachineList.back();
		mch->Reset();	
	}
}

/*---------------------------------------------------------------------------*
  Name:         PushStateMachine

  Description:  Pushes a state machine onto the front of the state machine list.

  Arguments:    mch : the new state machine
				initialize : whether to send the initial OnEnter event

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::PushStateMachine( StateMachine & mch, bool initialize )
{
	m_stateMachineList.push_back( &mch );
	
	if( initialize )
	{
		mch.Reset();
	}
}

/*---------------------------------------------------------------------------*
  Name:         PopStateMachine

  Description:  Pops a state machine from the front of the state machine list.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::PopStateMachine( void )
{
	ASSERTMSG( m_stateMachineList.size() > 1, "GameObject::PopStateMachine - Can't pop last state machine." );
	if( m_stateMachineList.size() > 1 ) {
		StateMachine * mch = m_stateMachineList.back();
		m_stateMachineList.pop_back();
		delete( mch );
		
		//Initialize new state machine
		mch = m_stateMachineList.back();
		mch->Reset();
	}
}

/*---------------------------------------------------------------------------*
  Name:         DestroyStateMachineList

  Description:  Destroys all state machines in the state machine list.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void GameObject::DestroyStateMachineList( void )
{
	while( m_stateMachineList.size() > 0 ) {
		StateMachine * mch = m_stateMachineList.back();
		m_stateMachineList.pop_back();
		delete( mch );
	}
}



